import pymysql

import approzium

from .._misc import patch
from .._mysql import MYSQLNativePassword, get_auth_resp


def _scramble_native_password(context, salt):
    # use normal method if not Approzium connection
    if not isinstance(context, dict):
        return pymysql._auth.scramble_native_password(context, salt)
    return get_auth_resp(
        context["authenticator"],
        context["host"],
        str(context["port"]),
        context["user"],
        MYSQLNativePassword,
        salt,
    )


class ApproziumConnection(pymysql.connections.Connection):
    def __init__(self, *args, authenticator=None, **kwargs):
        """Creates a PyMySQL connection through Approzium authentication. Takes
        the same arguments as ``pymysql.connect``, in addition to the
        authenticator argument.

        :param authenticator: AuthClient instance to be used for authentication. If
            not provided, the default AuthClient, if set, is used.
        :type authenticator: approzium.AuthClient, optional
        :raises: TypeError, if no AuthClient is given and no default one is set.
        :rtype: ``pymysql.connections.Connection``

        Example:

        .. code-block:: python

            >>> import approzium
            >>> from approzium.pymysql import connect
            >>> auth = approzium.AuthClient("authenticatorhost:6001", disable_tls=True)
            >>> con = connect("host=DB.com dbname=mydb", authenticator=auth)
            >>> # use the connection just like any other PyMySQL connection
        """
        if authenticator is None:
            authenticator = approzium.default_auth_client
        if authenticator is None:
            raise TypeError(
                "Auth client not specified and no default auth client is set"
            )
        self.authenticator = authenticator
        return super(ApproziumConnection, self).__init__(*args, **kwargs)

    @patch(pymysql._auth, "scramble_native_password", _scramble_native_password)
    def _request_authentication(self):
        # store info needed for Approzium authentication in password
        self.password = {
            "authenticator": self.authenticator,
            "host": self.host,
            "port": self.port,
            "user": self.user,
        }
        return super(ApproziumConnection, self)._request_authentication()



from functools import lru_cache

import boto3
import requests
from ec2_metadata import ec2_metadata


def assume_role(iam_role):
    sts_client = boto3.client("sts")
    response = sts_client.assume_role(
        DurationSeconds=3600, RoleArn=iam_role, RoleSessionName="Service1",
    )
    return response


def obtain_credentials(response):
    return response["Credentials"]


def obtain_claimed_arn(response):
    return response["AssumedRoleUser"]["Arn"]


def obtain_signed_get_caller_identity(credentials=None):
    if credentials is None:
        # Attempt to use local identity through the boto3 client.
        sts_client = boto3.client("sts")
    else:
        sts_session = boto3.Session(
            aws_access_key_id=credentials["AccessKeyId"],
            aws_secret_access_key=credentials["SecretAccessKey"],
            aws_session_token=credentials["SessionToken"],
        )
        sts_client = sts_session.client("sts")
    presigned_url = sts_client.generate_presigned_url("get_caller_identity", {})
    return presigned_url


def get_local_arn():
    sts_client = boto3.client("sts")
    response = sts_client.get_caller_identity()
    return response["Arn"]


@lru_cache(maxsize=1)
def is_ec2():
    # AWS docs reference:
    # https://docs.aws.amazon.com/AWSEC2/latest/UserGuide/identify_ec2_instances.html
    instance_identity_url = "http://169.254.169.254/latest/dynamic/instance-identity/"
    try:
        return requests.get(instance_identity_url, timeout=0.05).status_code == 200
    except requests.exceptions.ConnectionError:
        return False


def attribution_info():
    if not is_ec2():
        return {}
    attrs = ["public_hostname", "public_ipv4", "instance_id"]
    return {"ec2." + attr: getattr(ec2_metadata, attr) for attr in attrs}



import base64
import re

# try to import the secrets library from Python 3.6+ for the
# cryptographic token generator for generating nonces as part of SCRAM
# Otherwise fall back on os.urandom
try:
    from secrets import token_bytes as generate_token_bytes
except ImportError:
    from os import urandom as generate_token_bytes


class SCRAMAuthentication:
    AUTHENTICATION_METHODS = [b"SCRAM-SHA-256"]
    DEFAULT_CLIENT_NONCE_BYTES = 16  # 24
    REQUIREMENTS_CLIENT_FINAL_MESSAGE = ["client_channel_binding", "server_nonce"]
    REQUIREMENTS_CLIENT_PROOF = [
        "password_iterations",
        "password_salt",
        "server_first_message",
        "server_nonce",
    ]

    def __init__(self, authentication_method):
        self.authentication_method = authentication_method
        self.authorization_message = None
        # channel binding is turned off for the time being
        self.client_channel_binding = b"n,,"
        self.client_first_message_bare = None
        self.client_nonce = None
        self.client_proof = None
        self.password_salt = None
        self.password_iterations = None
        self.server_first_message = None
        self.server_key = None
        self.server_nonce = None

    def create_client_first_message(self, username):
        """Create the initial client message for SCRAM authentication"""
        self.client_nonce = self._generate_client_nonce(self.DEFAULT_CLIENT_NONCE_BYTES)
        # set the client first message bare here, as it's used in a later step
        self.client_first_message_bare = (
            b"n=" + username.encode("utf-8") + b",r=" + self.client_nonce
        )
        # put together the full message here
        msg = bytes()
        msg += self.authentication_method + b"\0"
        client_first_message = (
            self.client_channel_binding + self.client_first_message_bare
        )
        msg += (len(client_first_message)).to_bytes(
            4, byteorder="big"
        ) + client_first_message
        return msg

    def create_client_final_message(self, client_proof):
        """Create the final client message as part of SCRAM authentication"""
        if any(
            [
                getattr(self, val) is None
                for val in self.REQUIREMENTS_CLIENT_FINAL_MESSAGE
            ]
        ):
            raise Exception("you need values from server to generate a client proof")

        # generate the client proof
        msg = bytes()
        msg += (
            b"c="
            + base64.b64encode(self.client_channel_binding)
            + b",r="
            + self.server_nonce
            + b",p="
            + client_proof.encode("ascii")
        )
        return msg

    def parse_server_first_message(self, server_response):
        """Parse the response from the first message from the server"""
        self.server_first_message = server_response
        try:
            self.server_nonce = re.search(
                b"r=([^,]+),", self.server_first_message
            ).group(1)
        except IndexError:
            raise Exception("could not get nonce")
        if not self.server_nonce.startswith(self.client_nonce):
            raise Exception("invalid nonce")
        try:
            self.password_salt = re.search(
                b"s=([^,]+),", self.server_first_message
            ).group(1)
        except IndexError:
            raise Exception("could not get salt")
        try:
            self.password_iterations = int(
                re.search(b"i=(\d+),?", self.server_first_message).group(1)  # noqa:W605
            )
        except (IndexError, TypeError, ValueError):
            raise Exception("could not get iterations")

    def verify_server_final_message(self, server_final_message):
        """Verify the final message from the server"""
        try:
            server_signature = re.search(b"v=([^,]+)", server_final_message).group(1)
        except IndexError:
            raise Exception("could not get server signature")

        return server_signature == self.server_signature.encode("ascii")

    def _generate_client_nonce(self, num_bytes):
        token = generate_token_bytes(num_bytes)

        return base64.b64encode(token)

    def _generate_auth_msg(self):
        self.authorization_message = (
            self.client_first_message_bare
            + b","
            + self.server_first_message
            + b",c="
            + base64.b64encode(self.client_channel_binding)
            + b",r="
            + self.server_nonce
        )



import logging
import struct
from contextlib import contextmanager

logger = logging.getLogger(__name__)


def read_int32_from_bytes(bytes, index):
    num = struct.unpack("!i", bytes[index : index + 4])[0]
    return num


@contextmanager
def patch(module, attribute_name, new_value):
    original_value = getattr(module, attribute_name)
    setattr(module, attribute_name, new_value)
    try:
        yield
    finally:
        setattr(module, attribute_name, original_value)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: authenticator.proto
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import enum_type_wrapper

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='authenticator.proto',
  package='approzium.authenticator.protos',
  syntax='proto3',
  serialized_options=None,
  serialized_pb=b'\n\x13\x61uthenticator.proto\x12\x1e\x61pprozium.authenticator.protos\"J\n\x0b\x41WSIdentity\x12\"\n\x1asigned_get_caller_identity\x18\x01 \x01(\t\x12\x17\n\x0f\x63laimed_iam_arn\x18\x02 \x01(\t\"\xc4\x01\n\x0fPasswordRequest\x12G\n\x0f\x63lient_language\x18\x01 \x01(\x0e\x32..approzium.authenticator.protos.ClientLanguage\x12\x0e\n\x06\x64\x62host\x18\x02 \x01(\t\x12\x0e\n\x06\x64\x62port\x18\x03 \x01(\t\x12\x0e\n\x06\x64\x62user\x18\x04 \x01(\t\x12\x38\n\x03\x61ws\x18\x05 \x01(\x0b\x32+.approzium.authenticator.protos.AWSIdentity\"f\n\x10PGMD5HashRequest\x12\x44\n\x0bpwd_request\x18\x01 \x01(\x0b\x32/.approzium.authenticator.protos.PasswordRequest\x12\x0c\n\x04salt\x18\x02 \x01(\x0c\"\x99\x01\n\x13PGSHA256HashRequest\x12\x44\n\x0bpwd_request\x18\x01 \x01(\x0b\x32/.approzium.authenticator.protos.PasswordRequest\x12\x0c\n\x04salt\x18\x02 \x01(\t\x12\x12\n\niterations\x18\x03 \x01(\r\x12\x1a\n\x12\x61uthentication_msg\x18\x04 \x01(\t\"j\n\x14MYSQLSHA1HashRequest\x12\x44\n\x0bpwd_request\x18\x01 \x01(\x0b\x32/.approzium.authenticator.protos.PasswordRequest\x12\x0c\n\x04salt\x18\x02 \x01(\x0c\"0\n\rPGMD5Response\x12\x0c\n\x04hash\x18\x01 \x01(\t\x12\x11\n\trequestid\x18\x02 \x01(\t\"E\n\x10PGSHA256Response\x12\x0e\n\x06\x63proof\x18\x01 \x01(\t\x12\x0e\n\x06sproof\x18\x02 \x01(\t\x12\x11\n\trequestid\x18\x03 \x01(\t\"4\n\x11MYSQLSHA1Response\x12\x0c\n\x04hash\x18\x01 \x01(\x0c\x12\x11\n\trequestid\x18\x02 \x01(\t*?\n\x0e\x43lientLanguage\x12\x19\n\x15LANGUAGE_NOT_PROVIDED\x10\x00\x12\n\n\x06PYTHON\x10\x01\x12\x06\n\x02GO\x10\x02\x32\xfd\x02\n\rAuthenticator\x12q\n\x0cGetPGMD5Hash\x12\x30.approzium.authenticator.protos.PGMD5HashRequest\x1a-.approzium.authenticator.protos.PGMD5Response\"\x00\x12z\n\x0fGetPGSHA256Hash\x12\x33.approzium.authenticator.protos.PGSHA256HashRequest\x1a\x30.approzium.authenticator.protos.PGSHA256Response\"\x00\x12}\n\x10GetMYSQLSHA1Hash\x12\x34.approzium.authenticator.protos.MYSQLSHA1HashRequest\x1a\x31.approzium.authenticator.protos.MYSQLSHA1Response\"\x00\x62\x06proto3'
)

_CLIENTLANGUAGE = _descriptor.EnumDescriptor(
  name='ClientLanguage',
  full_name='approzium.authenticator.protos.ClientLanguage',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='LANGUAGE_NOT_PROVIDED', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='PYTHON', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='GO', index=2, number=2,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=873,
  serialized_end=936,
)
_sym_db.RegisterEnumDescriptor(_CLIENTLANGUAGE)

ClientLanguage = enum_type_wrapper.EnumTypeWrapper(_CLIENTLANGUAGE)
LANGUAGE_NOT_PROVIDED = 0
PYTHON = 1
GO = 2



_AWSIDENTITY = _descriptor.Descriptor(
  name='AWSIdentity',
  full_name='approzium.authenticator.protos.AWSIdentity',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='signed_get_caller_identity', full_name='approzium.authenticator.protos.AWSIdentity.signed_get_caller_identity', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='claimed_iam_arn', full_name='approzium.authenticator.protos.AWSIdentity.claimed_iam_arn', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=55,
  serialized_end=129,
)


_PASSWORDREQUEST = _descriptor.Descriptor(
  name='PasswordRequest',
  full_name='approzium.authenticator.protos.PasswordRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='client_language', full_name='approzium.authenticator.protos.PasswordRequest.client_language', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='dbhost', full_name='approzium.authenticator.protos.PasswordRequest.dbhost', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='dbport', full_name='approzium.authenticator.protos.PasswordRequest.dbport', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='dbuser', full_name='approzium.authenticator.protos.PasswordRequest.dbuser', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='aws', full_name='approzium.authenticator.protos.PasswordRequest.aws', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=132,
  serialized_end=328,
)


_PGMD5HASHREQUEST = _descriptor.Descriptor(
  name='PGMD5HashRequest',
  full_name='approzium.authenticator.protos.PGMD5HashRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='pwd_request', full_name='approzium.authenticator.protos.PGMD5HashRequest.pwd_request', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='salt', full_name='approzium.authenticator.protos.PGMD5HashRequest.salt', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=330,
  serialized_end=432,
)


_PGSHA256HASHREQUEST = _descriptor.Descriptor(
  name='PGSHA256HashRequest',
  full_name='approzium.authenticator.protos.PGSHA256HashRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='pwd_request', full_name='approzium.authenticator.protos.PGSHA256HashRequest.pwd_request', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='salt', full_name='approzium.authenticator.protos.PGSHA256HashRequest.salt', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='iterations', full_name='approzium.authenticator.protos.PGSHA256HashRequest.iterations', index=2,
      number=3, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='authentication_msg', full_name='approzium.authenticator.protos.PGSHA256HashRequest.authentication_msg', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=435,
  serialized_end=588,
)


_MYSQLSHA1HASHREQUEST = _descriptor.Descriptor(
  name='MYSQLSHA1HashRequest',
  full_name='approzium.authenticator.protos.MYSQLSHA1HashRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='pwd_request', full_name='approzium.authenticator.protos.MYSQLSHA1HashRequest.pwd_request', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='salt', full_name='approzium.authenticator.protos.MYSQLSHA1HashRequest.salt', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=590,
  serialized_end=696,
)


_PGMD5RESPONSE = _descriptor.Descriptor(
  name='PGMD5Response',
  full_name='approzium.authenticator.protos.PGMD5Response',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='hash', full_name='approzium.authenticator.protos.PGMD5Response.hash', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='requestid', full_name='approzium.authenticator.protos.PGMD5Response.requestid', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=698,
  serialized_end=746,
)


_PGSHA256RESPONSE = _descriptor.Descriptor(
  name='PGSHA256Response',
  full_name='approzium.authenticator.protos.PGSHA256Response',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='cproof', full_name='approzium.authenticator.protos.PGSHA256Response.cproof', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='sproof', full_name='approzium.authenticator.protos.PGSHA256Response.sproof', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='requestid', full_name='approzium.authenticator.protos.PGSHA256Response.requestid', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=748,
  serialized_end=817,
)


_MYSQLSHA1RESPONSE = _descriptor.Descriptor(
  name='MYSQLSHA1Response',
  full_name='approzium.authenticator.protos.MYSQLSHA1Response',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='hash', full_name='approzium.authenticator.protos.MYSQLSHA1Response.hash', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='requestid', full_name='approzium.authenticator.protos.MYSQLSHA1Response.requestid', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=819,
  serialized_end=871,
)

_PASSWORDREQUEST.fields_by_name['client_language'].enum_type = _CLIENTLANGUAGE
_PASSWORDREQUEST.fields_by_name['aws'].message_type = _AWSIDENTITY
_PGMD5HASHREQUEST.fields_by_name['pwd_request'].message_type = _PASSWORDREQUEST
_PGSHA256HASHREQUEST.fields_by_name['pwd_request'].message_type = _PASSWORDREQUEST
_MYSQLSHA1HASHREQUEST.fields_by_name['pwd_request'].message_type = _PASSWORDREQUEST
DESCRIPTOR.message_types_by_name['AWSIdentity'] = _AWSIDENTITY
DESCRIPTOR.message_types_by_name['PasswordRequest'] = _PASSWORDREQUEST
DESCRIPTOR.message_types_by_name['PGMD5HashRequest'] = _PGMD5HASHREQUEST
DESCRIPTOR.message_types_by_name['PGSHA256HashRequest'] = _PGSHA256HASHREQUEST
DESCRIPTOR.message_types_by_name['MYSQLSHA1HashRequest'] = _MYSQLSHA1HASHREQUEST
DESCRIPTOR.message_types_by_name['PGMD5Response'] = _PGMD5RESPONSE
DESCRIPTOR.message_types_by_name['PGSHA256Response'] = _PGSHA256RESPONSE
DESCRIPTOR.message_types_by_name['MYSQLSHA1Response'] = _MYSQLSHA1RESPONSE
DESCRIPTOR.enum_types_by_name['ClientLanguage'] = _CLIENTLANGUAGE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

AWSIdentity = _reflection.GeneratedProtocolMessageType('AWSIdentity', (_message.Message,), {
  'DESCRIPTOR' : _AWSIDENTITY,
  '__module__' : 'authenticator_pb2'
  # @@protoc_insertion_point(class_scope:approzium.authenticator.protos.AWSIdentity)
  })
_sym_db.RegisterMessage(AWSIdentity)

PasswordRequest = _reflection.GeneratedProtocolMessageType('PasswordRequest', (_message.Message,), {
  'DESCRIPTOR' : _PASSWORDREQUEST,
  '__module__' : 'authenticator_pb2'
  # @@protoc_insertion_point(class_scope:approzium.authenticator.protos.PasswordRequest)
  })
_sym_db.RegisterMessage(PasswordRequest)

PGMD5HashRequest = _reflection.GeneratedProtocolMessageType('PGMD5HashRequest', (_message.Message,), {
  'DESCRIPTOR' : _PGMD5HASHREQUEST,
  '__module__' : 'authenticator_pb2'
  # @@protoc_insertion_point(class_scope:approzium.authenticator.protos.PGMD5HashRequest)
  })
_sym_db.RegisterMessage(PGMD5HashRequest)

PGSHA256HashRequest = _reflection.GeneratedProtocolMessageType('PGSHA256HashRequest', (_message.Message,), {
  'DESCRIPTOR' : _PGSHA256HASHREQUEST,
  '__module__' : 'authenticator_pb2'
  # @@protoc_insertion_point(class_scope:approzium.authenticator.protos.PGSHA256HashRequest)
  })
_sym_db.RegisterMessage(PGSHA256HashRequest)

MYSQLSHA1HashRequest = _reflection.GeneratedProtocolMessageType('MYSQLSHA1HashRequest', (_message.Message,), {
  'DESCRIPTOR' : _MYSQLSHA1HASHREQUEST,
  '__module__' : 'authenticator_pb2'
  # @@protoc_insertion_point(class_scope:approzium.authenticator.protos.MYSQLSHA1HashRequest)
  })
_sym_db.RegisterMessage(MYSQLSHA1HashRequest)

PGMD5Response = _reflection.GeneratedProtocolMessageType('PGMD5Response', (_message.Message,), {
  'DESCRIPTOR' : _PGMD5RESPONSE,
  '__module__' : 'authenticator_pb2'
  # @@protoc_insertion_point(class_scope:approzium.authenticator.protos.PGMD5Response)
  })
_sym_db.RegisterMessage(PGMD5Response)

PGSHA256Response = _reflection.GeneratedProtocolMessageType('PGSHA256Response', (_message.Message,), {
  'DESCRIPTOR' : _PGSHA256RESPONSE,
  '__module__' : 'authenticator_pb2'
  # @@protoc_insertion_point(class_scope:approzium.authenticator.protos.PGSHA256Response)
  })
_sym_db.RegisterMessage(PGSHA256Response)

MYSQLSHA1Response = _reflection.GeneratedProtocolMessageType('MYSQLSHA1Response', (_message.Message,), {
  'DESCRIPTOR' : _MYSQLSHA1RESPONSE,
  '__module__' : 'authenticator_pb2'
  # @@protoc_insertion_point(class_scope:approzium.authenticator.protos.MYSQLSHA1Response)
  })
_sym_db.RegisterMessage(MYSQLSHA1Response)



_AUTHENTICATOR = _descriptor.ServiceDescriptor(
  name='Authenticator',
  full_name='approzium.authenticator.protos.Authenticator',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  serialized_start=939,
  serialized_end=1320,
  methods=[
  _descriptor.MethodDescriptor(
    name='GetPGMD5Hash',
    full_name='approzium.authenticator.protos.Authenticator.GetPGMD5Hash',
    index=0,
    containing_service=None,
    input_type=_PGMD5HASHREQUEST,
    output_type=_PGMD5RESPONSE,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='GetPGSHA256Hash',
    full_name='approzium.authenticator.protos.Authenticator.GetPGSHA256Hash',
    index=1,
    containing_service=None,
    input_type=_PGSHA256HASHREQUEST,
    output_type=_PGSHA256RESPONSE,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='GetMYSQLSHA1Hash',
    full_name='approzium.authenticator.protos.Authenticator.GetMYSQLSHA1Hash',
    index=2,
    containing_service=None,
    input_type=_MYSQLSHA1HASHREQUEST,
    output_type=_MYSQLSHA1RESPONSE,
    serialized_options=None,
  ),
])
_sym_db.RegisterServiceDescriptor(_AUTHENTICATOR)

DESCRIPTOR.services_by_name['Authenticator'] = _AUTHENTICATOR

# @@protoc_insertion_point(module_scope)



import logging

import asyncpg
from asyncpg.connect_utils import _ConnectionParameters
from asyncpg.protocol import Protocol

import approzium

from .._postgres import PGAuthClient, construct_msg, parse_msg

logger = logging.getLogger(__name__)


def appz_authenticate(self, data):
    # This method is used to override the method
    # `async.protocol.Protocol.data_received` and initiate and finish an
    # Approzium connection sequence. If the protocol instance is already
    # authenticated, this method reverts to the original `data_received` method

    context = self.password["approzium_context"]
    if context["authclient"] is None:
        # happens only once per connection
        context["authclient"] = PGAuthClient(
            None,  # read method is set on each run
            lambda msg: context["transport"].write(msg),
            context["authenticator"],
            context["host"],
            context["port"],
            self.user,
        )
    if not context["authclient"].done:
        # data could contain multiple message, in which case, we want AuthClient to
        # consume only the first message, while asyncpg consumes the rest
        msg_type, msg_content = parse_msg(data)
        first_msg_bytes = construct_msg(msg_type, msg_content)
        read_bytes = lambda: first_msg_bytes  # noqa: E731
        rest_bytes = data[len(first_msg_bytes) :]
        context["authclient"].read_bytes = read_bytes
        next(context["authclient"])
        if len(rest_bytes):
            Protocol.data_received(self, rest_bytes)
    else:
        Protocol.data_received(self, data)


def new_connection_made(self, transport):
    # `self` is a `asyncpg.protocol.Protocol` instance.

    # checks if this is an approzium connection
    if isinstance(self.password, dict) and "approzium_context" in self.password:
        # store transport because otherwise it is a private variable because of
        # Cython
        self.password["approzium_context"]["transport"] = transport

        # override the protocol's `data_received` method
        def new_data_received(data):
            appz_authenticate(self, data)

        self.data_received = new_data_received
    return original_connection_made(self, transport)


original_connection_made = Protocol.connection_made
Protocol.connection_made = new_connection_made


def new__connect_addr(*args, **kwargs):
    connection_class = kwargs["connection_class"]
    if connection_class.__name__ == "ApproziumConnection":
        host, port = kwargs["addr"]
        approzium_context = {
            "authenticator": connection_class.authenticator,
            "transport": None,  # this is determined later by `connection_made`
            "authclient": None,
            "host": host,
            "port": str(port),
        }
        # trick: put context in password field, discarding its passed value
        conn_params_dict = kwargs["params"]._asdict()
        conn_params_dict["password"] = {"approzium_context": approzium_context}
        new_conn_params = _ConnectionParameters(**conn_params_dict)
        kwargs["params"] = new_conn_params
    return original__connect_addr(*args, **kwargs)


original__connect_addr = asyncpg.connect_utils._connect_addr
asyncpg.connect_utils._connect_addr = new__connect_addr


def construct_approzium_connection(authenticator):
    # instantiate a fresh instance of the class on each connection
    # this instance is used to pass information that is needed during
    # connection process (e.g.: the authenticator)
    class ApproziumConnection(asyncpg.connection.Connection):
        pass

    ApproziumConnection.authenticator = authenticator
    return ApproziumConnection


async def connect(*args, authenticator=None, **kwargs):
    """Creates a Asyncpg connection through Approzium authentication. Takes
    the same arguments as ``asyncpg.connect``, in addition to the
    authenticator argument.

    :param authenticator: AuthClient instance to be used for authentication. If
        not provided, the default AuthClient, if set, is used.
    :type authenticator: approzium.AuthClient, optional
    :raises: TypeError, if no AuthClient is given and no default one is set.
    :rtype: ``asyncpg.Connection``

    Example:

    .. code-block:: python

        >>> import approzium
        >>> import asyncio
        >>> from approzium.asyncpg import connect
        >>> auth = approzium.AuthClient("myauthenticator.com:6001", disable_tls=True)
        >>> async def run():
        ...     con = await connect(user='postgres', authenticator=auth)
        ...     # use the connection just like any other Asyncpg connection
        ...     types = await con.fetch('SELECT * FROM pg_type')
        ...     print(types)
        >>> asyncio.get_event_loop().run_until_complete(run())
    """
    if authenticator is None:
        authenticator = approzium.default_auth_client
    conn = await asyncpg.connect(
        *args, **kwargs, connection_class=construct_approzium_connection(authenticator)
    )
    conn.authenticator = authenticator
    return conn



"""
Internal module that implements Approzium routines that are common to Postgres
libraries and the Postgres protocol.
"""
import logging
import struct

from .._misc import read_int32_from_bytes
from .scram import SCRAMAuthentication

logger = logging.getLogger(__name__)

# Postgres protocol constants
# derived from PGsource/src/include/libpq/pgcomm.h
AUTH_REQ_MD5 = 5
AUTH_REQ_SASL = 10


def parse_msg(msg):
    msg_type = msg[0]
    msg_size = struct.unpack("!i", msg[1 : 1 + 4])[0]
    msg_content = msg[5 : 5 + msg_size - 4]
    return msg_type, msg_content


def construct_msg(header, msg):
    if isinstance(header, int):
        header = header.to_bytes(1, "big")
    return header + struct.pack("!i", len(msg) + 4) + msg


def parse_auth_msg(msg):
    auth_type = read_int32_from_bytes(msg, 0)
    auth_info = {}
    if auth_type == AUTH_REQ_MD5:
        auth_info["salt"] = bytes(msg[-4:])
    elif auth_type == AUTH_REQ_SASL:
        if not msg[4:].startswith(b"SCRAM-SHA-256"):
            raise Exception("Server requested an unsupported SASL auth method")
        auth_info["scram"] = SCRAMAuthentication(b"SCRAM-SHA-256")
    return auth_type, auth_info


class PGAuthClient(object):
    """Class used to implement behaviour and state for an Approzium Postgres
    connection"""

    def __init__(self, read_bytes, write_bytes, authenticator, dbhost, dbport, dbuser):
        """When read_func is called with no arguments, a server message is
        received. When write_func is called with some bytes, they are written
        to the server.
        An instance should be instantiated when the next read_bytes call will
        return the first authentication request sent by the server."""
        self.read_bytes = read_bytes
        self.write_bytes = write_bytes
        self.authenticator = authenticator
        self.dbhost = dbhost
        self.dbport = dbport
        self.dbuser = dbuser
        self.done = False
        self.next_call = self.read_first_auth_msg

    def __next__(self):
        self.next_call()

    def read_first_auth_msg(self):
        msg_type, msg = parse_msg(self.read_bytes())
        if msg_type != ord("R"):
            raise Exception("Authentication message not received")
        auth_type, self.auth_info = parse_auth_msg(msg)
        if auth_type == AUTH_REQ_MD5:
            hash = self.authenticator._get_pg2_hash(
                self.dbhost,
                self.dbport,
                self.dbuser,
                AUTH_REQ_MD5,
                self.auth_info["salt"],
            )
            msg = construct_msg(b"p", b"md5" + hash.encode("ascii") + b"\0")
            self.write_bytes(msg)
            self.done = True
        elif auth_type == AUTH_REQ_SASL:
            scram_state = self.auth_info["scram"]
            client_first = scram_state.create_client_first_message(self.dbuser)
            msg = construct_msg(b"p", client_first)
            self.write_bytes(msg)
            self.next_call = self.scram_stage_2
        else:
            raise Exception("Unidentified authentication method")

    def scram_stage_2(self):
        msg_type, server_first = parse_msg(self.read_bytes())
        if msg_type != ord("R"):
            raise Exception("Error received unexpected response", server_first)
        # the part that is relevant is the part that starts with r=
        scram_state = self.auth_info["scram"]
        scram_state.parse_server_first_message(server_first[4:])
        client_final = self.authenticator._get_pg2_hash(
            self.dbhost,
            self.dbport,
            self.dbuser,
            AUTH_REQ_SASL,
            self.auth_info["scram"],
        )[0]
        msg = construct_msg(b"p", client_final)
        self.write_bytes(msg)
        self.next_call = self.scram_stage_3

    def scram_stage_3(self):
        msg_type, server_final = parse_msg(self.read_bytes())
        if msg_type != ord("R"):
            raise Exception("Error received unexpected response", server_final)
        scram_state = self.auth_info["scram"]
        if not scram_state.verify_server_final_message(server_final):
            raise Exception("Error bad server signature")
        self.done = True



from ._connect import ApproziumConnection


def connect(*args, **kwargs):
    return ApproziumConnection(*args, **kwargs)


connect.__doc__ = ApproziumConnection.__init__.__doc__
__all__ = ["connect"]



import opentelemetry.instrumentation.dbapi
from opentelemetry.instrumentation.dbapi import DatabaseApiIntegration

original_DatabaseApiIntegration = None


class _ApproziumDatabaseApiIntegration(DatabaseApiIntegration):
    def get_connection_attributes(self, connection):
        super().get_connection_attributes(connection)
        # check if this is an Approzium connection
        if hasattr(connection, "authenticator"):
            self.add_approzium_span_attributes(connection)

    def add_approzium_span_attributes(self, connection):
        for key, value in connection.authenticator.attribution_info.items():
            self.span_attributes["approzium." + key] = value


def instrument():
    """Instruments Opentelemetry ``opentelemetry.instrumentation.dbapi`` to automatically
    include attribution tags in traces generated by Approzium database connections.
    Call this method before calling the ``instrument`` method provided by Opentelemetry
    database extensions.

    Refer to :py:obj:`approzium.AuthClient.attribution_info` for generated tags.

    Example (using Opentelemetry Psycopg2 integration):

    .. code-block:: python

        >>> import approzium.opentelemetry
        >>> from approzium.psycopg2 import connect
        >>> from opentelemetry.instrumentation.psycopg2 import Psycopg2Instrumentor
        >>> approzium.opentelemetry.instrument()
        >>> Psycopg2Instrumentor().instrument()
        >>> cnx = connect("host=dbmd5 dbname=db user=bob")
        >>> cursor.execute("SELECT 1")
    """
    global original_DatabaseApiIntegration
    original_DatabaseApiIntegration = DatabaseApiIntegration
    opentelemetry.instrumentation.dbapi.DatabaseApiIntegration = (
        _ApproziumDatabaseApiIntegration
    )


def uninstrument():
    """Removes Approzium instrumentation from Opentelemetry
    ``opentelemetry.instrumentation.dbapi``

    :raises: Exception, if ``opentelemetry.instrumentation.dbapi`` was not instrumented
    by Approzium
    """
    if original_DatabaseApiIntegration is None:
        raise Exception("Opentelemetry is not currently instrumented by Approzium")
    opentelemetry.instrumentation.dbapi.DatabaseApiIntegration = (
        original_DatabaseApiIntegration
    )
